import React, { ReactNode } from 'react';
import { Column, ConditionalRowStyle } from '../SearchUI/types';
import './DataTable.css';
export interface DataTableProps {
    /**
     * The ID used to identify this component in Dash callbacks
     */
    id?: string;
    /**
     * Dash-assigned callback that should be called whenever any of the
     * properties change
     */
    setProps?: (value: any) => any;
    /**
     * Class name(s) to append to the component's default class.
     * By default, the .box and .p-0 classes are applied to the table, but these will be
     * overridden if a className is supplied.
     */
    className?: string;
    /**
     * Array of data objects to display in the table
     */
    data: any[];
    /**
     * An array of column definition objects to control what is rendered in the table.
     * See `Column` documentation for specifics on how to construct `Column` objects.
     */
    columns?: Column[];
    /**
     * Optionally include a field to sort by on initial load
     * Must be a valid field and included in your list of columns
     */
    sortField?: string;
    /**
     * If including a sortField, set whether it should ascend by default
     * True for ascending, False for descending
     */
    sortAscending?: boolean;
    /**
     * Optionally include a secondary sort field. If the sortField ever becomes the same as
     * the secondarySortField, the secondary field is removed.
     * Must be a valid field and included in your list of columns.
     */
    secondarySortField?: string;
    /**
     * If including a secondarySortField, set whether it should ascend by default.
     * True for ascending, False for descending.
     */
    secondarySortAscending?: boolean;
    /**
     * List of conditions for styling rows based on a property (selector) and a value.
     * Note that this prop currently only supports checking for
     * value equivalence (i.e. row[selector] === value).
     * See `ConditionalRowStyle` documentation for how to construct `ConditionalRowStyle` conditions.
     */
    conditionalRowStyles?: ConditionalRowStyle[];
    /**
     * Optionally include/exclude checkboxes next to rows for selecting
     */
    selectableRows?: boolean;
    /**
     * Property to maintain the state of selected rows so that
     * they are accessible via Dash callback
     */
    selectedRows?: any[];
    /**
     * Combine with selectableRows prop to only allow one row to be selected at a time.
     */
    singleSelectableRows?: boolean;
    /**
     * Set to true to show a header with total number of rows and a columns selector
     */
    hasHeader?: boolean;
    /**
     * Optional class name to apply to the table header
     */
    headerClassName?: string;
    /**
     * A noun in singular form to describe what a result represents (e.g. "material").
     * This is displayed in the table header.
     */
    resultLabel?: string;
    /**
     * Plural form of the result label. If none supplied, it will automatically be the result label plus an "s"
     */
    resultLabelPlural?: string;
    /**
     * Set to true to paginate the table records
     */
    pagination?: boolean;
    /**
     * If true, an expanded component will be used for pagination (same as in `SearchUI`).
     * If false, a compact version will be used.
     */
    paginationIsExpanded?: boolean;
    /**
     * Content to display below the table but inside the table's box wrapper.
     * Accepts markdown.
     */
    footer?: ReactNode;
    /**
     * This is a temporary solution to allow SearchUI's to render in Storybook.
     * There is an issue with the dynamic column header components that causes
     * Storybook to crash. Rendering column headers as plain strings fixes the problem.
     * Note that this will disable column tooltips and unit labels.
     */
    disableRichColumnHeaders?: boolean;
}
/**
 * Component for rendering data in a table.
 * Uses react-data-table-component under the hood.
 */
export declare const DataTable: React.FC<DataTableProps>;
