import { PeriodicTableSelectionMode } from '../../periodic-table/PeriodicTableModeSwitcher/PeriodicTableModeSwitcher';
import { MaterialsInputType } from '../MaterialsInput';
export declare const VALID_ELEMENTS: string[];
export declare const ELEMENTS_REGEX: RegExp;
export declare const ELEMENTS_SPLIT_REGEX: RegExp;
export declare const parseFormula: (formula: any) => string[] | null;
/**
 * Validates a string as a checmical formula and returns
 * a parsed array of valid elements if successful.
 *
 * @param {string} formula An unparsed chemical formula string
 * @param {RegExp} illegalCharsRegex Optional param to override the default regex for finding illegal characters
 * @param {RegExp} elementsRegex Optional param to override the default regex for finding element symbols
 * @returns {string[] or undefined} Array of valid element symbols or undefined
 */
export declare const validateFormula: (formula: string, illegalCharsRegex?: RegExp, elementsRegex?: RegExp) => string[] | undefined;
/**
 * Validates a string as a chemical formula that allows
 * ranged and wildcard (x) subscripts.
 *
 * @param {string} formula An unparsed chemical formula string
 * @returns {string[] or undefined} Array of valid element symbols or undefined
 */
export declare const validateRangedFormula: (formula: string) => string[] | undefined;
/**
 * Determines if a string is delimited by commas, hyphens, or spaces
 * If multiple delimiters are present, the delimiter with the lowest index is used
 * If none is present, defaults to comma
 *
 * @param {string} input String to parse for a delimiter
 * @returns Delimiter as a regular expression object
 */
export declare const getDelimiter: (input: string) => RegExp;
/**
 * Validates a string as a valid list of element symbols or wildcards.
 * Elements can be delimited by a comma, space, or hyphen
 *
 * @param {string} elementStr String of element symbols to be validated
 * @returns Array of valid element symbols
 */
export declare const validateElements: (elementStr: string, delimiter?: RegExp | undefined) => string[] | undefined;
export declare const validateElementsList: (elementStr: string) => string[] | undefined;
export declare const validateChemicalSystem: (elementStr: string) => string[] | undefined;
/**
 * Validates a string as a possible Simplified Molecular-Input Line-Entry System (SMILES)
 *
 * @param {string} value String value to be validated
 * @returns Validated SMILES string or null
 */
export declare const validateSmiles: (value: string) => string | null;
/**
 * Validates a string as a possible material, molecule, or battery id
 * Note: this only validates the string's format, not its existence in the database.
 *
 * @param {string} value String value to be validated
 * @returns Validated id string or null
 */
export declare const validateMPID: (value: string) => string | null;
/**
 * Validate the number of elements in an input's parsed value.
 * If the input type is not an element-like type, the length will always be valid.
 * @param parsedValue either a validated input string or a validated list of elements
 * @param type the input type that the parsedValue came from
 * @param maxElements maximum number of allowable elements
 * @returns true or false
 */
export declare const validateInputLength: (parsedValue: string | string[] | undefined, type: MaterialsInputType | null, maxElements?: number | undefined) => boolean;
export declare type MaterialsInputTypesMap = Partial<Record<MaterialsInputType, any>>;
/**
 * Object to map MaterialsInputType values to other values
 * such as validation functions or periodic table selection modes.
 * Object keys must be one of the values defined in the MaterialsInputType enum.
 */
export declare const materialsInputTypes: MaterialsInputTypesMap;
export declare const getMaterialsInputTypeByMappedValue: (key: string, value: any) => MaterialsInputType | undefined;
/**
 * Validate a MaterialsInput value.
 * This method will always return an array with two values.
 * Failed validations will return null in the second value in the returned array.
 *
 * @param {string} value Input value string for a MaterialsInput
 * @param {string} inputType [optional] MaterialsInputType type to use to validate the input value (will detect field type if not included)
 * @returns Array with two values:
 *   1. The MaterialsInputType (only null if no field is supplied or detected)
 *   2. The validated parsed value returned from the field's validation method (null if validation failed)
 */
export declare const validateInputType: (value: string, inputType?: MaterialsInputType | undefined) => [MaterialsInputType | null, any];
/**
 * Validate a MaterialsInput value.
 * This method will always return an array with two values.
 * Failed validations will return null in the second value in the returned array.
 *
 * @param {string} value Input value string for a MaterialsInput
 * @param {string} allowedInputTypes [optional] MaterialsInputType type to use to validate the input value (will detect field type if not included)
 * @returns Array with two values:
 *   1. The MaterialsInputType (only null if no field is supplied or detected)
 *   2. The validated parsed value returned from the field's validation method (null if validation failed)
 */
export declare const detectAndValidateInputType: (value: string, allowedInputTypes: MaterialsInputType[]) => [MaterialsInputType | null, any];
/**
 * Map the list of allowed input types to a list of allowed periodic table selection modes.
 * This prevents periodic table modes dropdown from having items that are
 * inconsistent with the allowed input types.
 * The order that these items are appended determines the order they are rendered in the periodic table.
 */
export declare const getAllowedSelectionModes: (allowedInputTypes: MaterialsInputType[]) => PeriodicTableSelectionMode[];
