import React, { ReactNode } from 'react';
import './CrystalToolkitScene.less';
import { ExportType } from '../scene/constants';
import { CameraState } from '../CameraContextProvider/camera-reducer';
export interface CrystalToolkitSceneProps {
    /**
     * The ID used to identify this component in Dash callbacks
     */
    id?: string;
    /**
     * Dash-assigned callback that should be called whenever any of the
     * properties change
     */
    setProps?: (value: any) => any;
    /**
     * First child will be rendered as the settings panel.
     * Second child will be rendered as the bottom panel (legend).
     */
    children?: ReactNode;
    /**
     * Class name that will wrap around the whole scene component.
     * When enlarged, this class name is applied to the modal-content element.
     */
    className?: string;
    /**
     * Add a debugging view
     */
    debug?: boolean;
    /**
     * Scene JSON, the easiest way to generate this is to use the Scene class
     * in crystal_toolkit.core.scene and its to_json method.
     */
    data: any;
    /**
     * Options used for generating scene.
     * Supported options and their defaults are given as follows:
     * {
     *    antialias: true, // set to false to improve performance
     *    renderer: 'webgl', // 'svg' also an option, used for unit testing
     *    transparentBackground: false, // transparent background
     *    background: '#ffffff', // background color if not transparent,
     *    sphereSegments: 32, // decrease to improve performance
     *    cylinderSegments: 16, // decrease to improve performance
     *    staticScene: true, // disable if animation required
     *    defaultZoom: 1, // 1 will zoom to fit object exactly, <1 will add padding between object and box bounds
     *    zoomToFit2D: false // if true, will zoom to fit object only along the X and Y axes (not Z)
     *    extractAxis: false // will remove the axis from the main scene
     * }
     * There are several additional options used for debugging and testing,
     * please consult the source code directly for these.
     */
    settings?: any;
    /**
     * Hide/show nodes in scene by its name (key), value is 1 to show the node
     * and 0 to hide it.
     */
    toggleVisibility?: any;
    /**
     * File type to be downloaded as an image. Either png or dae
     */
    imageType?: ExportType;
    /**
     * THIS PROP IS SET AUTOMATICALLY
     * Data string for the image generated on image button click
     */
    imageData?: string;
    /**
     * THIS PROP IS SET AUTOMATICALLY
     * Date string that represents the time imageData was set.
     * Use this prop in dash callbacks to trigger downloads of imageData.
     */
    imageDataTimestamp?: any;
    /**
     * List of options to show in file download dropdown
     */
    fileOptions?: string[];
    /**
     * THIS PROP IS SET AUTOMATICALLY
     * The last file type clicked in the file download menu
     */
    fileType?: string;
    /**
     * THIS PROP IS SET AUTOMATICALLY
     * Date string that represents the time fileType was set.
     * Use this prop in dash callbacks to trigger file downloads.
     */
    fileTimestamp?: any;
    onObjectClicked?: (value: any) => any;
    /**
     * Size of axis inlet
     */
    inletSize?: number;
    /**
     * Size of scene
     */
    sceneSize?: number | string;
    /**
     * Padding of axis inlet
     */
    inletPadding?: number;
    /**
     * Orientation of axis view
     */
    axisView?: string;
    /**
     * Animation
     *
     * Set up animation styles
     *
     * 'play'
     * 'none'
     * 'slider'
     */
    animation?: string;
    /**
     * THIS PROP IS SET AUTOMATICALLY
     * Object that maintains the current state of the camera.
     * e.g.
     * {
     *   position: {x: 0, y: 0, z: 0},
     *   quarternion: {x: 0, y: 0, z: 0, w: 0},
     *   zoom: 1,
     *   setByComponentId: "1",
     *   following: true
     * }
     */
    currentCameraState?: CameraState;
    /**
     * Object for setting the scene to a custom camera state.
     * When modified, the camera will update to new custom state.
     * e.g.
     * {
     *   position: {x: 0, y: 0, z: 0},
     *   quarternion: {x: 0, y: 0, z: 0, w: 0}, (optional)
     *   zoom: 1 (optional)
     * }
     */
    customCameraState?: CameraState;
    /**
     * Determines if control bar is visible
     * @default true
     */
    showControls?: boolean;
    /**
     * Determines if expand button in control bar is visible
     * @default true
     */
    showExpandButton?: boolean;
    /**
     * Determines if image download button in control bar is visible
     * @default true
     */
    showImageButton?: boolean;
    /**
     * Determines if file export button in control bar is visible
     * @default true
     */
    showExportButton?: boolean;
    /**
     * Determines if return to original position button in control bar is visible
     * @default true
     */
    showPositionButton?: boolean;
}
/**
 * This component is intended to draw simple 3D scenes using the popular
 * Three.js scene graph library. In particular, the JSON representing the 3D scene
 * is intended to be human-readable, and easily generated via Python. This is not
 * intended to be a replacement for a full scene graph library, but for rapid
 * prototyping by non-experts.
 */
export declare const CrystalToolkitScene: React.FC<CrystalToolkitSceneProps>;
